/****************************************************************************
*
*   Copyright (c) 2009 www.roborealm.com
*   (please link back if you use this code!)
*
*   This program is free software; you can redistribute it and/or modify
*   it under the terms of the GNU General Public License version 2 as
*   published by the Free Software Foundation.
*
*   Alternatively, this software may be distributed under the terms of BSD
*   license.
*
****************************************************************************/

#include "SoR_Utils.h" //includes all the technical stuff

/*
****************************************************************************
*
*   This file includes a simple servo controller program that will listen to
*   the UART for incoming commands to adjust any one of the 29 servos possible
*   in the Axon processor.
*
*	Note that the Axon processor has fewer than 29 timers and thus special timing
*	considerations are needed in order to create 29 PWM signals without interrupts.
*
*	The servo timing requirement requires that the system not be interrupted during
*	the critical first phase timing of servos otherwise servo cliches will occur.
*	Because of this it is possible to lose incoming serial characters esp with a
*	fast PC writing values faster than the Axon can handle given the timing 
*	restrictions. Because of this loss the protocol uses a sync bit (128) to
*	ensure that successive signals are aligned correctly, uses binary values
*	to reduce the serial communication as much as possible, and includes a CRC
*	to ensure that the data received can be confirmed as being valid.
*
*	The timing of the servos needs to be done in a staggered fashion. You cannot
* 	turn on the servo, wait and then turn it off as with 29 servos you will quickly
*	exceed the 20ms base period of most servos. Thus the technique instead will
*	turn on all servos at once and then independantly turn off those servos in
*	succession based on their timing period. For example, if you have two seros
*	at 1ms and 1.5ms phase you would turn both on, wait for 1ms, turn off the first
*	servo, then wait 0.5ms and finally turn off the second. This requires the timing
* 	to be sorted in an ascending manner such that the delay from one servo to the
*	next can be easily calculated. While ideally this would be implemented in a
*	parallel manner the short delay due to a single thread is acceptible to most
*	servos.
*
***************************************************************************


Protocol:

identification
	1st byte - command id = 128

servo_on
	1st byte - command id = 129
	2nd byte - servo number
	3rd byte - CRC

servo_off
	1st byte - command id = 130
	2nd byte - servo number
	3rd byte - CRC

set_servo_position
	1st byte - command id = 131
	2nd byte - servo number
	3rd byte - servo position (bits 1-4)
	4th byte - servo position (bits 5-8)
	5th byte - CRC for last 4 bytes

All bytes need to be < 128 such that the 5th byte can also be used as a sync byte in case a byte
is lost during transmission.

If two sync bytes are seen Axon responds with the word Axon in order for the PC to know that it is
communicating correctly.

***************************************************************************

*/

// servo position prior to sorting
int init_servo_position[29]={-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1,-1};
// base servo number
int init_servo_number[29]={6,5,4,3,2,7,6,5,4,3,2,1,0,1,2,3,4,5,6,7,6,7,6,5,4,3,2,1,0}; 
// base servo port
int *init_servo_port[29]={&PORTH,&PORTH,&PORTH,&PORTH,&PORTH,&PORTE,&PORTE,&PORTE,&PORTE,&PORTE,&PORTE,&PORTE,&PORTA,&PORTA,&PORTA,&PORTA,&PORTA,&PORTA,&PORTA,&PORTA,&PORTJ,&PORTC,&PORTC,&PORTC,&PORTC,&PORTC,&PORTC,&PORTC,&PORTC};

// sorted servo position
int servo_position[29];
// sorted servo number
int servo_number[29];
// sorted servo port
unsigned int *servo_port[29];

// cmd variables are used to maintain state for incoming characters
// such that the above protocol can be implemented.
int cmd_index = 0;
int cmd_id;
int cmd_servo_number;
int cmd_servo_position_1;
int cmd_servo_position_2;
int cmd_crc;

// defines the timing period of min and max servo values. You may need to adjust these for your servos
#define MIN_SERVO_TIME 275
#define MAX_SERVO_TIME 1150
float fact = ((MAX_SERVO_TIME-MIN_SERVO_TIME)/255.0);

// is called when a new character from the PC is available
void get_data(unsigned char c)
{
	cli();
	//rprintf("%c", c);

	// toggle LED
	PORTB ^= (1<<6);

	// check for our sync byte ... which is also the id
	if (c&128)
	{
		cmd_index=0;
		cmd_id = (int)c;
	}

	// id 128 is use for the PC to know that this is an Axon program
	if (cmd_id==128)
	{
		rprintf("Axon\n");
	}
	else
	// enable or disble a servo. When a servo is disabled it can be manually moved.
	if ((cmd_id==129)||(cmd_id==130))
	{
		if (cmd_index==1)
		{
			cmd_servo_number=c;
		}
		else
		if (cmd_index==2)
		{
			cmd_crc=c;

			if (((cmd_id^(cmd_servo_number<<1))&127)==cmd_crc)
			{
				if (cmd_id==129)
					init_servo_position[cmd_servo_number]=MIN_SERVO_TIME+(((MAX_SERVO_TIME-MIN_SERVO_TIME)/255)*128);
				else
					init_servo_position[cmd_servo_number]=-1;
			}
		}
	}
	else
	// used to set a servo position. Note that you need 4 bytes of info for this.
	if (cmd_id==131)
	{
		if (cmd_index==1)
			cmd_servo_number=c;
		else
		if (cmd_index==2)
			cmd_servo_position_1=c;
		else
		if (cmd_index==3)
			cmd_servo_position_2=c;
		else
		if (cmd_index==4)
		{
			cmd_crc = (int)c;
			// CRC is a simple XOR with each successive byte shifted by one more bit
			if (((cmd_id^(cmd_servo_number<<1)^(cmd_servo_position_1<<2)^(cmd_servo_position_2<<3))&127)==cmd_crc)
			{
				// construct the actual value ... we cannot use 1 byte for this since
				// a servo value ranges from 0-255 which would cause it to be interpreted
				// as a sync byte for values above 128. Thus we need to use two bytes
				// to contain half the information.
				int pos = cmd_servo_position_1|(cmd_servo_position_2<<4);
				init_servo_position[cmd_servo_number]=MIN_SERVO_TIME+(fact*pos);
			}
		}
	}

	// increment our data counter
	cmd_index++;
}

int main(void)
{
	int i,j;

	uartInit();  // initialize the UART (serial port)
    uartSetBaudRate(0, 38400); // set UARTE speed, for Bluetooth
    uartSetBaudRate(1, 115200); // set UARTD speed, for USB connection
    uartSetBaudRate(2, 38400); // set UARTH speed
    uartSetBaudRate(3, 38400); // set UARTJ speed, for Blackfin

	rprintfInit(uart1SendByte);// initialize rprintf system and configure uart1 (USB) for rprintf

	configure_ports(); 

	timer0Init();

	uartSetRxHandler(1, &get_data);

	//initialize ISR:
	UCSR1B |= (1 << RXCIE1); // Enable the USART Recieve Complete interrupt (USART_RXCn)
	sei();//turn on the Global Interrupt Enable flag

	while (1)
	{
		// copy all initial values prior to sorting
		memcpy(servo_position, init_servo_position, sizeof(servo_position));
		memcpy(servo_port, init_servo_port, sizeof(servo_port));
		memcpy(servo_number, init_servo_number, sizeof(servo_number));

		// sort based on position information and keep track of number and port
		// we're using a quick and dirty bubble sort which for 29 values should
		// be ok. You could instead use an insertion sort which might be a little
		// quicker per loop iteration as we don't expect the position to change
		// frequently, but with the (N*N)/2 and flag optimizations this should
		// function fine.
		int isDone=0;
		for (i=0;(i<29-1)&&(!isDone);i++)
		{
			isDone=1;
			for (j=0;j<29-i-1;j++)
			{
				if (servo_position[j]>servo_position[j+1])
				{
					// swap servo position 
					int tmp = servo_position[j];
					servo_position[j] = servo_position[j+1];
					servo_position[j+1] = tmp;

					// keep number and port in sync with the position so we know which port to change!
					tmp = servo_number[j];
					servo_number[j] = servo_number[j+1];
					servo_number[j+1] = tmp;

					tmp = servo_port[j];
					servo_port[j] = servo_port[j+1];
					servo_port[j+1] = tmp;

					isDone=0;
				}
			}
		}

		//turn off interrupts .. this is very important otherwise your servo will jitter!
		cli();

		// set all the ports to on that are enabled (i.e.>0)
		for (i=0;i<29;i++)
		{
			if (servo_position[i]>0)
			{
				PORT_ON(*servo_port[i], servo_number[i]);
			}
		}

		// delay until the last position has occured
		int offset=0;
		for (i=0;i<29;i++)
		{
			if (servo_position[i]>0)
			{
				if (servo_position[i]>offset)
				{
					// wait until this servo's position offset
					delay_cycles(servo_position[i]-offset);
					offset=servo_position[i];
				}
				// turn off this servo and continue with next servos
				PORT_OFF(*servo_port[i], servo_number[i]);
			}
		}

		// renable interrupts so that we can receive UART/serial values
		sei();

		// wait for a total of 20ms from the last servo timing value.
		timerPause((4666-offset)/234);
	}
}
